from pathlib import Path
from typing import Union

from lxml import etree
from lxml.html import HtmlElement

from .parser import get_html_parser

TreeOrErrMessage = Union[HtmlElement, str]


def get_html_as_tree(path_to_html_file: Path, custom_element=None) -> TreeOrErrMessage:
    """
    Lower Level Interface - Retrieve Tree from File
    """

    parser = get_html_parser(custom_element=HtmlElement) if custom_element is None \
        else get_html_parser(custom_element=custom_element)

    if path_to_html_file.exists():
        return etree.parse(path_to_html_file.as_uri(), parser=parser).getroot()
    else:
        return "There is no HTML File at the given path location"


def persist_html_tree(html_tree: HtmlElement, path_to_html_file: Path) -> None:
    """
    Lower Level Interface - Persist a given HTML Tree safely at a given file path
    """
    dir_path = path_to_html_file.parent

    if not dir_path.exists():
        dir_path.mkdir(parents=True, exist_ok=True)

    with path_to_html_file.open(mode='wb') as f:
        f.write(etree.tostring(html_tree, pretty_print=True, method="html"))



import unicodedata
from typing import List

from lxml.etree import _ElementTree
from lxml.html import HtmlElement


class NoHTMLElementFound(Exception):
    def __init__(self, exception_msg):
        self.exception_msg = exception_msg


class MultipleHTMLElementsFound(Exception):
    def __init__(self, exception_msg):
        self.exception_msg = exception_msg


class HTMLUtility:

    @staticmethod
    def get_html_element_by_custom_id(current_tree: HtmlElement,
                                      html_attr_name: str,
                                      custom_element_id: int) -> HtmlElement:
        """
        Focused on unique identifiers
        Note: could even be abstracted to find elements given <any_value> for <any_html_attr>
        :param current_tree:  the current HTML Page In Memory representation
        :param html_attr_name: The custom attr name
        :param custom_element_id: An Identifier for an instance in the custom Namespace
        :rtype: HtmlElement
        """

        pseudo_xpath = "//*[@" + html_attr_name + "='{ELEMENT_ID}']".format(
            ELEMENT_ID=custom_element_id
        )
        found_elements: List[HtmlElement] = current_tree.xpath(pseudo_xpath)

        if len(found_elements) == 0:
            raise NoHTMLElementFound("There is no element for {ATTR_NAME}={ELEMENT_ID} in the current HTML Tree".format(
                ATTR_NAME=html_attr_name,
                ELEMENT_ID=custom_element_id
            ))
        elif len(found_elements) > 1:
            raise MultipleHTMLElementsFound(
                "Inconsistency: There are multiple elements found for {ATTR_NAME}={ELEMENT_ID}. "
                "But the result should be one unique element".format(
                    ATTR_NAME=html_attr_name,
                    ELEMENT_ID=custom_element_id
                ))
        else:
            return found_elements[0]

    @staticmethod
    def get_html_element_by_attr_value_combi(current_tree: HtmlElement, html_attr_name: str,
                                             html_attr_value: str) -> List[HtmlElement]:
        """

        Args:
            current_tree: The current HTML Page In Memory representation
            html_attr_name: an attr name
            html_attr_value: an attr value

        Returns:
            A list of html elements with the specified attr:value combination

        """
        pseudo_xpath = "//*[@" + html_attr_name + "='{HTML_ATTR_VALUE}']".format(
            HTML_ATTR_VALUE=html_attr_value
        )
        found_elements: List[HtmlElement] = current_tree.xpath(pseudo_xpath)
        return found_elements

    @staticmethod
    def construct_canonical_xpath(current_tree: HtmlElement, html_attr_name: str, custom_element_id: int) -> str:
        """
        :param current_tree:  the current HTML Page In Memory representation
        :param html_attr_name: the custom attr name
        :param custom_element_id: An Identifier for an instance in the custom Namestance
        :rtype: HtmlElement
        """
        root_tree: _ElementTree = current_tree.getroottree()
        target_element = HTMLUtility.get_html_element_by_custom_id(current_tree, html_attr_name, custom_element_id)
        return root_tree.getpath(target_element)

    @staticmethod
    def collect_text_content(element: HtmlElement) -> str:
        """
        Collect all text content associated with an Element safely

        Accounts for the following constellation <parent> <child>text_content</child>tail_content</parent>
        Where the child.tail holds the tail_content
        """
        element_text = element.text if hasattr(element, "text") else ""
        element_text = "" if element_text is None else element_text

        element_tail = element.tail if hasattr(element, "tail") else ""
        element_tail = "" if element_tail is None else element_tail

        element_text_content = element_text + " " + element_tail
        element_text_content = unicodedata.normalize('NFKC', element_text_content)
        element_text_content.rstrip()

        return element_text_content

    @staticmethod
    def check_only_children_of_one_kind(element: HtmlElement) -> bool:
        children = element.getchildren()
        if len(children) == 0:
            return False  # children of None kind not of one kind

        tag_to_be = children[0].tag
        for child in children:
            if child.tag != tag_to_be:
                return False
        return True

    @staticmethod
    def check_has_multiple_childs(element: HtmlElement) -> bool:
        no_of_children = len(element.getchildren())
        return no_of_children > 2



from lxml import etree
from lxml.html import HtmlElement


def get_html_parser(custom_element=HtmlElement):
    """
    Returns an HTML Parser configured to an custom_element, if provided
    :param custom_element: an Element drived from lxml ElementBase Class
    :type custom_element: Subclass of etree.ElementBase
    :return: an HTML Parser Object
    :rtype: lxml.etree.HTMLParser
    """

    parser = etree.HTMLParser(encoding="utf-8", remove_comments=True, remove_blank_text=True, remove_pis=True)
    if custom_element is not None:
        parser_lookup = etree.ElementDefaultClassLookup(element=custom_element)
        parser.set_element_class_lookup(parser_lookup)
    return parser



import inspect
from typing import Callable, List, Tuple, Union

from lxml.html import HtmlElement

# Type Definitions
P = Union[HtmlElement, None]
D = Union[dict, None]

OP = Union[  # for Operation Functions
    List[Callable[[HtmlElement, int, D], None]],
    Tuple[Callable[[HtmlElement, int, D], None]],
    Callable[[HtmlElement, int, D], None]
]
CO = Callable[[HtmlElement], int]  # for Child Order Functions


def higher_order_traversal(
        operations: OP,
        child_order: CO = None,
        depth_first=True,
        pre_order=True
) -> Callable[[HtmlElement], Tuple[int, dict]]:
    """
    Abstraction over Traversal Action
    Higher Order Traversal Function that returns customized traversal functions
    :param operations: Tuple of Functions that that perform Operations on the current Node
    :type operations: operations: (func(node, counter))
    :param child_order: Determine Order in which childrens are visited
    :type child_order: func(node) -> int
    :param depth_first:
    :type depth_first:
    :param pre_order: Only meaningful for depth first traversal
           Indicate whether Node (PreOrder) or children (PostOrder) are processed first
    :type: bool
    :return: Customized Traversal Function
    :rtype: Callable[[HtmlElement], Callable]
    """

    def tag_traversal(node: HtmlElement,
                      ops: OP = operations,
                      c_order: CO = child_order,
                      counter: int = 0,
                      collection_dict={}
                      ) -> Tuple[int, dict]:
        """
        :param node: see above
        :type node: HtmlElement
        :param ops: see above
        :type ops: OP
        :param c_order: see above
        :type c_order: CO
        :param counter: Counter
        :type counter: int
        :param collection_dict: Dictionary to Collect Information
        :type collection_dict: dict
        :return: No of tree Elements, updated Collection Dictionary
        :rtype: Tuple[int, dict]
        """

        def process_node(n: HtmlElement,
                         o: OP,
                         c: int,
                         coll: D) -> int:
            """
            :param n: node
            :param o: operations
            :param c: counter
            :param coll: collection_dictionary
            :return updated Counter
            """
            if n.tag is not None:
                if type(o) is not tuple:
                    o = tuple(o) if type(o) is list else (o,)
                    for operation in o:
                        sig = inspect.signature(operation)

                        # Decision whether CollectionDict Updating Operation
                        if 'collection_dict' in sig.parameters:
                            operation(n, c, coll)
                        else:
                            operation(n, c)
                    c += 1
                    # print("Processeing Node")
                    # print("CounterUpdate: " + str(c))
                    return c

        # Actual Traversal
        inner_counter = counter  # Collect Counter Updates within single node

        if depth_first:
            if pre_order:
                inner_counter = process_node(node, ops, inner_counter, collection_dict)

            # Process Children
            if len(node):
                children: list = node.getchildren()
                if c_order is not None:
                    children.sort(key=c_order)

                for child in children:
                    inner_counter, collection_dict = tag_traversal(
                        node=child,
                        counter=inner_counter,
                        collection_dict=collection_dict)

            if not pre_order:  # i.e Post Order
                inner_counter = process_node(node, ops, inner_counter, collection_dict)

        if not depth_first:
            current_level: List[HtmlElement] = [node]
            while len(current_level) > 0:
                next_level = []
                for current_node in current_level:
                    inner_counter = process_node(current_node,
                                                 ops,
                                                 inner_counter,
                                                 collection_dict)

                    # Collect next level Nodes
                    if len(current_node):
                        children: list = current_node.getchildren()
                        if c_order is not None:
                            children.sort(key=c_order)

                        for child in children:
                            next_level.append(child)

                current_level = next_level

        return inner_counter, collection_dict

    return tag_traversal


def empty_tags(node: HtmlElement, counter: int):
    """
    Empties the JSS, Style contents
    """
    if node.get("style"):
        node.set("style", "")

    if node.tag == "script":
        node.text = ""
    elif node.tag == "style":
        node.text = ""
        return
    else:
        return


def count_visit_operation(node: HtmlElement, counter: int):
    visit_counter = int(node.get("visit_counter", 0))
    visit_counter += 1
    node.set("visit_counter", str(visit_counter))
    return


print_operations = [
    lambda node, counter: print(node.tag)
]


def track_tags_operation(node: HtmlElement, counter: int, collection_dict: dict):
    collection_dict['tag-sequence'].append(node.tag)


track_operations = print_operations + [track_tags_operation]

# Customized Traversal Function, ready to be loaded from other modules
simple_track_traversal_pre_order = higher_order_traversal(track_operations)
simple_track_traversal_post_order = higher_order_traversal(track_operations, pre_order=False)
simple_track_traversal_breadth_first = higher_order_traversal(track_operations, depth_first=False)



from time import sleep

import dpp_common_utils.selenium
import requests
from lxml import etree
from lxml.html import HtmlElement
from selenium.common.exceptions import WebDriverException

from html_processing.parser import get_html_parser


def lxml_connect(url, custom_element=None):
    """
    Connects to URL and returns the HTML Tree
    :param url: The URL to the target data source
    :type url: str
    :return: a Subclass of etree.ElementBase, dependent on what is specified in the parser
    :rtype: HtmlElement
    """

    parser = get_html_parser() if custom_element is None else get_html_parser(custom_element)
    driver = dpp_common_utils.selenium.get_driver(True)
    __check_url(url)
    try:
        driver.get(url)
    except WebDriverException:
        raise requests.exceptions.ConnectionError

    sleep(3)  # sleep - the script has to be loaded and this needs time
    res = driver.page_source

    html_tree: HtmlElement = etree.fromstring(res, parser=parser)
    return html_tree


def __check_url(url):
    """
    Sends a request to an URL to check if it's valid and throws Errors if it's not
    Uses snippet from https://stackoverflow.com/questions/62024144/validate-urls-using-python-and-selenium


    :param url: The URL to check
    :type url: str
    :throws: requests.exceptions
    """

    req = requests.get(url)  # This already throws InvalidURL, InvalidSchema etc. if given String is not valid
    if req.status_code != requests.codes['ok']:
        raise requests.exceptions.HTTPError()



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

with open('VERSION') as version_file:
    version = version_file.read().strip()

setuptools.setup(
    name="html_processing",
    version=version,
    author="J. Hohenstein",
    author_email="j.hoh@t-online.de",
    description="Helps processing html with lxml",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://git.uni-due.de/",
    packages=setuptools.find_packages(exclude=("tests",)),
    install_requires=[
        'lxml~=4.5.1',
        'requests~=2.24.0',
        'dpp_common_utils~=0.2.0',
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7',
)
